(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Edoardo Biagioni (Edoardo.Biagioni@cs.cmu.edu)
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Ken Cline (Kenneth.Cline@cs.cmu.edu)
        Nick Haines (Nick.Haines@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

		i.	Abstract

	Support for parsing and viewing locatives.

	Given a description of what is wanted and bytes to view (in the
	form of a ubyte1 flat_vector), forge parses the bytes, builds
	a descriptor, and locviews the data.  This is type unsafe and
	ugly, but necessary in order to use current flat data.


		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	functor Forge
	2.	datatype forge_desc
	3.	local fun descriptor_makestring
	4.	fun forge_descriptor
	5.	fun forge

		iii.	RCS Log
	
$Log: forge.fun,v $
Revision 1.2  1993/12/21  17:15:37  cline
Cleaned up following code review.

Revision 1.1  1993/12/06  16:31:40  cline
Initial revision



		1.	functor Forge
*)

local
  structure D  = System.Descriptors
  structure L  = Locative
  structure FV = Flat_Vector
in
  functor Forge () : FORGE =
  struct

(*
		2.	datatype forge_desc
 *)
    datatype forge_desc =
      Forge_Byte1
    | Forge_Byte2
    | Forge_Byte4
    | Forge_Byte2u		(* unaligned *)
    | Forge_Byte4u		(* unaligned *)
    | Forge_Tuple of forge_desc list
				(* tuple of listed forge_descs *)
    | Forge_Vector of descriptor
				(* vector containing the rest of the data *)
    | Forge_Vector_Length of descriptor*int
				(* vector of n bytes *)
    | Forge_CString
				(* null terminated ubyte1 flat_vector *)

   (* Raise forge_desc when the data can't be parsed as described *)
   exception Forge_Desc

   (*
     Unaligned numeric types are not primitive.  To determine what
     descriptors they use, we build each item (using from_int), and
     query the system for the descriptor.
    *)
   val desc_byte2u = D.desc_of (Unaligned.Byte2u.from_int 0)
   val desc_byte4u = D.desc_of (Unaligned.Byte4u.from_int 0)
      

(*
	3.	local fun descriptor_makestring

	This exists for debugging.  It does not belong here!
 *)
   val descriptor_makestring =
     let
       fun m desc_byte1 = "desc_byte1"
         | m desc_byte2 = "desc_byte2"
         | m desc_byte4 = "desc_byte4"
         | m (desc_tuple dvec) =
             let fun loop k = if k < Vector.length dvec then
                                (if k>0 then ", " else "") ^
                                m (Vector.sub (dvec, k)) ^
                                loop (k+1)
                             else ""
             in
               "desc_tuple [" ^ loop 0 ^ "]"
             end
         | m (desc_vector (n, d)) =
             "desc_vector (" ^ (Integer.makestring n) ^ ", " ^ m d ^ ")"
         | m desc_box = "desc_box"
     in
       m
     end

(*
		4.	fun forge_descriptor

	forge_descriptor parses flat according to fdesc and
	returns a descriptor for the desired view.
 *)

   fun forge_descriptor fdesc flat =
     let
(*       val desc_vector (len, _) = L.locdesc flat*)
       val len = case L.locdesc flat of
                   desc_vector (l, _) => l
                 | _ => raise Forge_Desc

       fun find_zero k =
         if k >= len then
           ( print "(forge.fun)find_zero: no zero to be found!\n";
             raise Forge_Desc
           )
         else if L.locget (L.locsub (flat, k)) = 1u0 then k
         else find_zero (k+1)

       (* parse starting at pos k according to fd (forge desc).
          return desc and next position *)
       fun parse_from (Forge_Byte1, k) = (desc_byte1, k+1)
         | parse_from (Forge_Byte2, k) = (desc_byte2, k+2)
         | parse_from (Forge_Byte4, k) = (desc_byte4, k+4)
         | parse_from (Forge_Byte2u, k) = (desc_byte2u, k+2)
         | parse_from (Forge_Byte4u, k) = (desc_byte4u, k+4)
         | parse_from (Forge_Tuple fdlist, k) =
             let
               fun f (fd, (dlist, l)) =
                 case parse_from (fd, l) of (d, l') => (d::dlist, l')
             in
               case revfold f fdlist ([], k) of
                 (dlist, l) => (desc_tuple (Vector.vector (rev dlist)), l)
             end
         | parse_from (Forge_Vector_Length (desc, n), k) =
             (desc_vector (n, desc), k + (n * (D.desc_size_of desc)))
         (* vector filling the rest of the data *)
         | parse_from (Forge_Vector (desc), k) =
             parse_from (Forge_Vector_Length
                           (desc, (len - k) div (D.desc_size_of desc)),
                  k)
         | parse_from (Forge_CString, k) =
             let
               val l = (find_zero k) + 1
             in
               (desc_vector (l-k, desc_byte1), l)
             end
     in
       case parse_from (fdesc, 0) of
         (desc, k) => if k <> len
                        then
                           (
                             print ("(forge.sml)forge_descriptor: length mismatch ("^(makestring k)^"<>"^(makestring len)^")\n");
                             raise Forge_Desc
                           )
                        else desc
     end

(*
		5.	fun forge
 *)

   fun forge (fdesc:forge_desc) (flat:ubyte1 flat_vector locative) =
     L.locview (flat, forge_descriptor fdesc flat)

  end (* functor Forge *)
end (* local *)
