(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Edoardo Biagioni (Edoardo.Biagioni@cs.cmu.edu)
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Ken Cline (Kenneth.Cline@cs.cmu.edu)
        Nick Haines (Nick.Haines@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

		i.	Abstract



		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	functor kinit
	2.	 Instantiate Kerberos functors
	3.	fun kinit
	4.	fun export_kinit

		iii.	RCS Log
	
$Log: kinit.fun,v $
Revision 1.2  1994/05/24  14:26:34  robby
Ken's resturcturing, following a code review.

Revision 1.1  1993/12/14  21:12:13  cline
Initial revision


		1.	functor kinit
*)

functor Kinit (structure B       : FOX_BASIS
               structure Convert : CONVERT
               val do_prints     : bool
               val get_afs_ticket: bool): KINIT =
struct

(*
		2.	 Instantiate Kerberos functors
 *)

  structure Realm = Krb_Realm(val config_file = "/etc/krb.conf")

  structure Protocol = Krb_Protocol(structure Realm = Realm)

  structure Tkt_File = Krb_Tkt_File(structure B        = B
                                    structure Protocol = Protocol
                                    structure Convert  = Convert)

  structure Kname_Parse  = Kname_Parse(structure B        = B
                                       structure Protocol = Protocol)

  structure DES = DES(val dflag = false)

  structure Swap_Bytes = Swap_Bytes ()

  structure Forge = Forge ()

  structure KDC = Krb_KDC(structure B = B
                          structure Protocol = Protocol
                          structure Realm    = Realm
                          structure Convert = Convert
                          val max_retries = 20
                          val sleep_time  = 100
                          val udp_header_size = 42
                          val ip_over_eth = 2ux800
                          val udp_over_ip = 1u17
                          val udp_local_port = 2ux5742
                          val do_prints   = do_prints)

  structure Util = Krb_Util(structure B = Fox_Basis
                            structure Protocol = Protocol
                            structure KDC = KDC
                            structure DES = DES
                            structure Convert = Convert
                            structure Swap_Bytes = Swap_Bytes
                            structure Forge = Forge
                            val do_prints = do_prints)

  type credential = Protocol.credential

  structure AFS =
  struct
    val maxktcnamelen = 64
    val maxktcrealmlen = 64
    fun squash (s, len) =
          {| string = Convert.string2flat_vector (s^"\000"),
             filler = Flat_Vector.flat_vector_const (1u0, len - (size s + 1))
           |}
    fun ktc_principal (Protocol.Principal {name, instance, realm}) =
          {| name     = squash (name,     maxktcnamelen),
             instance = squash (instance, maxktcnamelen),
             realm    = squash (realm,    maxktcrealmlen) |}
  end

(*
		3.	fun kinit
 *)
  fun kinit (principal_string:string, lifetime:int) =
    let
      val realm = Realm.local_realm ()
      val principal =
          let
            val Protocol.Principal
                  {name=pname, instance=pinstance, realm=prealm} =
                     Kname_Parse.parse principal_string
          in
            Protocol.Principal
              { name     = pname,
                instance = pinstance,
                realm    = if prealm <> "" then prealm else realm }
          end
      val _ = B.V.print "Kinit: Getting ticket granting ticket\n"
      val tg_credential =		(* ticket granting credential *)
          Util.get_tgticket
            { principal = principal,
              service   = Protocol.ticket_granting_service,
              life      = Byte1.from_int lifetime }

      val _ = B.V.print "Kinit: Getting afs credential\n"
      val afs_credential =
            if get_afs_ticket then
              SOME (Util.request
                      { principal     = principal,
                        service       = Protocol.Local_Principal
                                          {name = "afs", instance = ""},
                        srealm        = realm,
                        checksum      = 4u0,
                        lifetime      = 4u200,
                        tg_credential = tg_credential })
            else
              NONE
    in
      Tkt_File.clear_ticketfile
        (case principal of
           Protocol.Principal {name, instance, ...} =>
             Protocol.Local_Principal {name=name, instance=instance});
      Tkt_File.append_ticketfile tg_credential;
      case afs_credential of
        SOME (Protocol.Credential cr) =>
          let
            val afs = AFS.ktc_principal
                        (Protocol.Principal
                           {name="afs", instance="", realm=realm})
            val me  = AFS.ktc_principal principal
            val ticket = case #ticket cr of
                           Protocol.Ticket tkt => tkt
            val key    = case #session_key cr of
                           Protocol.Key k => k
            val ticket_length = Flat_Vector.flat_length ticket
            fun life_to_time (start, life) =
                  start + (Byte4.from_int (300*(Byte1.to_int life)))
            val token =
                  {| start_time  = #issue_date cr,
                     end_time    = life_to_time
                                     ((#issue_date cr), (#lifetime cr)),
                     session_key = key,
                     kvno        = Byte2.from_int (Byte1.to_int (#kvno cr)),
                     ticket_len  = Unaligned.Byte4u.unalign
                                     (Byte4.from_int ticket_length),
                     ticket      = ticket,
                     filler      = Flat_Vector.flat_vector_const
                                     (1u0, 344 - ticket_length) |}
            val set_afs_token =
                  (System.Unsafe.CInterface.c_function "afs_set_token") :
                    (ByteArray.bytearray *
                     ByteArray.bytearray * 
                     ByteArray.bytearray) -> unit
            fun flat2ba x =
                  Convert.flat_vector2bytearray
                   (Locative.locget (Convert.flatten (locative x)))
          in
            Tkt_File.append_ticketfile (Protocol.Credential cr);
            set_afs_token (flat2ba afs, flat2ba token, flat2ba me)
          end
      | NONE => ()
    end
end

(*
	4.	fun export_kinit

	Create a heap called "kinit-heap", which run kinit when executed.
 *)
local
  structure Kinit = Kinit ( structure B = Fox_Basis
                            structure Convert = Convert()
                            val do_prints = false
                            val get_afs_ticket = true)
in
  fun export_kinit () =
    let
      fun run (argv, environ) =
       (
        case argv of
          [_, _, _, principal] => Kinit.kinit (principal, 200)
        | _ => print "Usage: kinit <principal>\n"
       )
       handle System.Unsafe.CInterface.SysError (x, s) =>
                    (print ("SysError: " ^ s ^ "\n");
                     raise (System.Unsafe.CInterface.SysError (x, s)))
            | x => raise x
    in
      exportFn ("kinit-heap", run)
    end (* let *)
end (* local *)
