(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Edoardo Biagioni (Edoardo.Biagioni@cs.cmu.edu)
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Ken Cline (Kenneth.Cline@cs.cmu.edu)
        Nick Haines (Nick.Haines@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

		i.	Abstract

	A real-life test for IP options.

		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	functor Test_Ip_Options_Real

		iii.	RCS Log
	
$Log: ipoptreal.tst,v $
# Revision 1.1  1995/06/20  17:05:46  esb
# Initial revision
#

		1.	functor Test_Ip_Options_Real

	Build all three types of Time_Stamps and a record route and
	bounce them off a known ICMP echo host as a weak test of
	functionality.  This is the only reasonable way to test this
	functionality as ICMP strict routing is often not implemented
	(and is not on the CMU gateway).

	For example, I aim an ICMP echo through GW.CS.CMU.EDU to
	CMU_DEFAULT.NET.CMU.EDU with a strict source route. If the
	echo returns my data, then the packet arrived and its options
	did not cause it to break.

	To run this on your own system, use a program like traceroute
	to show you a path of two gateways. Use a program like ping to
	test that the second host is accepting and returning ICMP
	messages. Then instantiate the structure with the appropriate
	ip addresses.

	These tests should not be run automatically.

*)

functor Test_Ip_Options_Real (structure B: FOX_BASIS
			       sharing type B.Filter.device_filter
				          = ByteArray.bytearray
			      structure A: ADDRESSING
			      val gateway1: FoxWord32.word
			      val gateway2: FoxWord32.word
			      val debug_level: int ref option
			      val here: FoxWord32.word): TEST_STRUCTURE =
 struct
  structure Trace = Trace (structure V = B.V
			   val debug_level = debug_level
			   val module_name = "ipoptreal.tst")
  val local_print = Trace.local_print
  val debug_print = Trace.debug_print 

  val ip_over_eth = Numbers.n2ux800

  structure Dev = Build_Eth_Dev (structure B = B
				 val high_priority_filter = true)

  structure Ip_Stack = Build_Ip (structure Device = Dev.Eth_Dev
				 structure B = B
				 val ip_protocol = ip_over_eth)

  fun make_byte4 (high, mid1, mid2, low) =
       let val join = FoxWord32.orb
           val shift = FoxWord32.lshift
           fun convert n = FoxWord32.intToWord n
       in join (join (shift (convert high, 24), shift (convert mid1, 16)),
	        join (shift (convert mid2, 8), convert low))
       end

  structure Ip = Ip_Stack.Ip

  fun run_test (name, options) =
       (B.Format.byte_base := B.Format.Hex;
	Ip.initialize ();
	let val queue = B.Event_Queue.new ()
	    fun receive (Ip.Ip_Packet {data, ...}) =
	         let val size = B.Dyn_Array.size data
		     val format =
		          B.Format.format_bytearray [B.Format.Bytes size]
		     val formats = format (B.Dyn_Array.read data, 0)
		 in local_print ("received " ^ fold op^ formats "");
		    B.Test.test (name, fn _ => true);
		    B.Event_Queue.signal {queue = queue, match = fn _ => true,
					  value = ()};
		    ()
		 end
	    fun status (Ip.Option_Packet (options, _)) =
	         local_print ("received options " ^
			      Ip.Option.makestring_options options)
	      | status s =
	         local_print ("received status " ^ Ip.makestring_status s)
	    fun handler _ = (receive, status)
	    val address = Ip.Address {ip = gateway1, proto = Numbers.n1u1}
	    val connection = Ip.connect (address, Ip.Handler handler)
	    val icmp_data =
	         let val b = ByteArray.array (15, 0)
		     val u1 = FoxWord8.update
		     val u2 = FoxWord16.update
		     val sum = B.Checksum.checksum
		     val compl = B.Checksum.one_s_complement
		     val length = ByteArray.length b
		     val big = B.Order.B2.to_big
		 in (* As per RFC 792. *)
		    u1 (b, 0, Numbers.n1u8); (* Echo Request Message. *)
		    u1 (b, 1, Numbers.n1u0); (* code = 0, always *)
		    u2 (b, 2, Numbers.n2u0); (* checksum *)
		    u2 (b, 4, big Numbers.n2u1); (* identifier one *)
		    u2 (b, 6, big Numbers.n2u2); (* sequence two *)
		    u1 (b, 8, Numbers.n1u0); (* data... *)
		    u1 (b, 9, Numbers.n1u1);
		    u1 (b, 10, Numbers.n1u2);
		    u1 (b, 11, Numbers.n1u3);
		    u1 (b, 12, Numbers.n1u4);
		    u1 (b, 13, Numbers.n1u5);
		    u1 (b, 14, Numbers.n1u6);
		    u2 (b, 2, big (compl (sum (b, 0, length))));
		    b
		 end
	    fun send_fun connection () =
	         let val size = ByteArray.length icmp_data
		     val (packet, send) = Ip.allocate_send (connection, size)
		 in B.Dyn_Array.update (packet, 0, icmp_data);
		    send ()
		 end
	    val sleep_time = if Trace.debug_on () then 10000 else 10000
	    fun timeout () =
	         (B.Scheduler.sleep sleep_time;
		  case B.Event_Queue.signal {queue = queue,
					     match = fn _ => true,
					     value = ()} of
		     NONE => ()
		   | SOME _ => B.Test.test (name, fn _ => false))
	    fun while_waiting () =
	         (B.Scheduler.fork timeout;
		  B.Scheduler.fork (send_fun connection))
	in Ip.set_options (connection, options);
	   B.Event_Queue.wait {queue = queue, event = (),
			       while_waiting = while_waiting};
	   Ip.close connection;
	   Ip.finalize ();
	   ()
	end
       handle _ => (Ip.finalize (); ()))

  fun run_tests () =
       let val one_step = {length = Numbers.n1u7, position = Numbers.n1u4,
			   route = [B.Order.B4.to_big here]}
           val one_step = {length = Numbers.n1u15, position = Numbers.n1u4,
			   route = [here, gateway1, here]}
	   val two_free = {length = Numbers.n1u11, position = Numbers.n1u4,
			   route = [B.Order.B4.to_big gateway1]}
	   val three_free = {length = Numbers.n1u11, position = Numbers.n1u4,
			     route = []}
	   val gate2_route = {length = Numbers.n1u7, position = Numbers.n1u4,
			      route = [B.Order.B4.to_big gateway2]}
	   val gate2_route = {length = Numbers.n1u11, position = Numbers.n1u4,
			      route = [gateway1, gateway2]}
	   val time_stamp1 = {length = Numbers.n1u12, position = Numbers.n1u5,
			      overflow = Numbers.n1u0,
			      stamps = Ip.Option.Stamp []}
	   val time_stamp2 = {length = Numbers.n1u12, position = Numbers.n1u5,
			      overflow = Numbers.n1u0,
			      stamps = Ip.Option.Stamp []}
	   val time_stamp3 = {length = Numbers.n1u20, position = Numbers.n1u5,
			      overflow = Numbers.n1u1,
			      stamps = Ip.Option.Stamp []}
	   val time_stamp4 = {length = Numbers.n1u20, position = Numbers.n1u5,
			      overflow = Numbers.n1u3,
			      stamps = Ip.Option.Stamp []}
       in run_test ("no options", []);
	  run_test ("strict source routing",
		    [Ip.Option.Strict_Route one_step]);
  (* Send a Time_Stamp to see if our gateway even reads our Option. *)
	  run_test ("time stamp 1", [Ip.Option.Time_Stamp time_stamp1]);
  (* Send it through to gateway 2. *)
	  run_test ("strict source routing",
		    [Ip.Option.Strict_Route gate2_route]);
	  run_test ("time stamp 2", [Ip.Option.Strict_Route two_free,
				     Ip.Option.Time_Stamp time_stamp2]);
	  run_test ("time stamp 3", [Ip.Option.Strict_Route two_free,
				     Ip.Option.Time_Stamp time_stamp3]);
	  run_test ("time stamp 4", [Ip.Option.Strict_Route two_free,
				     Ip.Option.Time_Stamp time_stamp4]);
	  run_test ("loose source routing", [Ip.Option.Loose_Route two_free]);
	  run_test ("record route", [Ip.Option.Strict_Route two_free,
				    Ip.Option.Record_Route three_free]);
	  ()
       end (* let *)

  exception Initialization_Error

  fun set_interfaces [] = ()
    | set_interfaces ((A.Interface {name, ip, mask, gateways}) :: rest) =
       (Ip_Stack.Ip_Mux.add (name, ip);
	Ip.initialize ();
	Ip.set_interface_address (name, ip);
	case mask of
	   NONE => ()
	 | SOME m => Ip.set_subnet_mask (name, SOME m);
	case gateways of
	   [] => ()
	 | (A.Gateway {name, ip} :: others) => Ip.set_default_gateway ip)

  fun init_ip () =
       (A.initialize (! A.default);
	case A.fetch () of
	   NONE => raise Initialization_Error
	 | SOME (A.Host {name, interfaces}) =>
	    set_interfaces interfaces)

  fun run () =
       (init_ip ();
	B.Test.tests ("IP-options-real", 9, run_tests);
	Ip.finalize ();
	())

 end (* struct *)

(*
		1.	structure Test_Ip_Options_Real
*)

structure Test_Ip_Options_Real =
    Test_Ip_Options_Real (structure B = Fox_Basis
			  structure A = Addressing
			  val debug_level = NONE
			  val here = Test_Addresses.get_ip "sly"
			  val gateway1 = Test_Addresses.get_ip "gw"
			  val gateway2 = Test_Addresses.get_ip "cmu-default")

