(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Edo Biagioni (esb@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

	i.	Abstract

	cache.fun: functor Cache which, given a key type, a value type,
	and a hash function on the keys, produces a SINGLE_CACHE structure.

	ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	functor Single_Cache

	iii.	RCS Log
$Log:	cache.fun,v $
Revision 1.3  97/06/04  11:44:35  esb
added optional tracing.

Revision 1.2  97/04/22  11:32:05  esb
changed to remove a key before adding.

Revision 1.1  97/04/17  14:06:20  esb
Initial revision


	1.	functor Single_Cache

	This cache is implemented as an array of pairs of (key, value)
	entries.  In each pair, the most-recently accessed element is
	to the left, so that if we need to remove an entry from the
	cache, we remove the right element.  (Appropriate allowances
	are made for array elements that contain zero or one entries
	instead of two).
*)

functor Single_Cache (structure V: VENDOR
		      type key
		      type value
		      val max_elements: int
		      val max_space: int
		      val eq: key * key -> bool
		      val hash: key -> word
		      val debug_level: int ref option): SINGLE_CACHE =
 struct
  exception Insufficient_Space

  local
   structure Trace = Trace (structure V = V
			    val debug_level = debug_level
			    val module_name = "cache.fun"
			    val makestring = fn _ => NONE)

   type entry = {key: key, value: value, space: int}
   (* note: when "Two", the left entry is the one more recently accessed. *)
   datatype element = Empty | One of entry | Two of entry * entry

   val entries = Int.max (max_elements, 0) div 2 + 1
   val cache = Array.array (entries, Empty)
   val cache_space = ref 0

(* divide the hash by two to make sure that the conversion to int does
   not raise overflow *)
   fun index key = (Word.toInt ((hash key) div 0w2)) mod entries

   fun string_array index =
        if index < entries then
	 (case Array.sub (cache, index) of
	     Empty => (Int.toString index ^ ": [] == ")
	   | One _ => (Int.toString index ^ ": one == ")
	   | Two _ => (Int.toString index ^ ": two == ")) ^
	  string_array (index + 1)
	else ""
        
   fun flush_loop 0 = cache_space := 0
     | flush_loop n =
	(Array.update (cache, n - 1, Empty);
	 flush_loop (n - 1))

   fun flush_older_loop 0 = ()
     | flush_older_loop n =
        (case Array.sub (cache, n - 1) of
	    Empty => ()
	  | One {key, value, space} => ()
	  | Two (new, old as {key, value, space}) =>
	     (Array.update (cache, n - 1, One new);
	      cache_space := ! cache_space - space);
	 flush_older_loop (n - 1))

   fun add_space space =
        (if space > max_space then
	  (Trace.debug_print (fn _ => "insufficient space to add " ^
			      Int.toString space ^ ", max space is " ^
			      Int.toString max_space);
	   false)
	 else
	  (if max_space - space < ! cache_space then
	    (Trace.trace_constant_string "flushing older entries";
	     flush_older_loop entries;
	     if max_space - space < ! cache_space then
	      (Trace.trace_constant_string "flushing entire cache";
	       flush_loop entries)
	     else ())
	   else ();
	   cache_space := ! cache_space + space;
	   Trace.debug_print (fn _ => "cache space is " ^
			      Int.toString (! cache_space) ^
			      " after adding " ^ Int.toString space);
	   true))

   fun element_loop (result, f, n) =
        if n >= entries then result
	else
	 case Array.sub (cache, n) of
	    Empty => element_loop (result, f, n + 1)
	  | One {key, value, space} =>
	     element_loop (f ((key, value), result), f, n + 1)
	  | Two ({key = new_key, value = new_value, space = new_space},
		 {key = old_key, value = old_value, space = old_space}) =>
	     element_loop (f ((new_key, new_value),
			      f ((old_key, old_value), result)), f, n + 1)

   fun makestring_pair ("", string, separator) = string
     | makestring_pair (first, new, separator) = first ^ separator ^ new

   fun makestring_entry (make, separator) (entry, head) =
	makestring_pair (head, make entry, separator)

  in

   type key = key
   type value = value
   type space = int

  fun remove key =
        (case Array.sub (cache, index key) of
	    Empty => ()
	  | One (entry as {key = entry_key, value, space}) =>
	     if eq (key, entry_key) then
	      (Array.update (cache, index key, Empty);
	       cache_space := ! cache_space - space)
	     else ()
	  | Two (new as {key = new_key, value = _, space = new_space},
		 old as {key = old_key, value = _, space = old_space}) =>
	     if eq (key, new_key) then
	      (Array.update (cache, index key, One old);
	       cache_space := ! cache_space - new_space)
	     else if eq (key, old_key) then
	       (Array.update (cache, index key, One new);
		cache_space := ! cache_space - old_space)
	     else ())
	  handle x => Trace.print_handled (x, SOME "remove")

   fun add (key, value, space) =
        (remove key;			(* remove it if it is present *)
         if add_space space then
          (let val index = index key
	       val entry = {key = key, value = value, space = space}
	   in case Array.sub (cache, index) of
	         Empty => Array.update (cache, index, One entry)
	       | One old =>
		  Array.update (cache, index, Two (entry, old))
	       | Two (newer, older) =>	(* drop the older entry *)
		  Array.update (cache, index, Two (entry, newer))
	   end)
	    handle x => Trace.print_raise_again (x, SOME "add")
	 else raise Insufficient_Space)

  fun look key =
        (case Array.sub (cache, index key) of
	    Empty => NONE
	  | One (entry as {key = entry_key, value, space}) =>
	     if eq (key, entry_key) then SOME value else NONE
	  | Two (new as {key = new_key, value = new_value, space = _},
		 old as {key = old_key, value = old_value, space = _}) =>
	     if eq (key, new_key) then SOME new_value
	     else if eq (key, old_key) then
	      (* re-order entries. *)
	      (Array.update (cache, index key, Two (old, new));
	       SOME old_value)
	     else NONE)
	  handle x => (Trace.print_handled (x, SOME "look"); NONE)

   fun flush () = flush_loop entries

   fun app f =
        element_loop ((), fn (entry, ()) => (f entry; ()), 0)

   fun fold f init =
        element_loop (init, f, 0)

   fun makestring (string, separator) =
        element_loop ("", makestring_entry (string, separator), 0)

  end
 end
